# Copyright 2025 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO(yashkatariya): Remove this after NamedSharding supports more complicated
# shardings like sub-axes, strided shardings, etc.
from jax._src.lib import xla_client
from jax._src.sharding_impls import GSPMDSharding as GSPMDSharding


def get_op_sharding_from_serialized_proto(
    sharding: bytes) -> xla_client.OpSharding:
  proto = xla_client.OpSharding()
  proto.ParseFromString(sharding)
  return proto


def get_hlo_sharding_from_serialized_proto(
    sharding: bytes) -> xla_client.HloSharding:
  return xla_client.HloSharding.from_proto(
      get_op_sharding_from_serialized_proto(sharding))


def get_serialized_proto_from_hlo_sharding(
    sharding: xla_client.HloSharding) -> bytes:
  return sharding.to_proto().SerializeToString()
